Left

Struct Left 

pub struct Left<A, B, const TOTAL: u64 = 2> { /* private fields */ }
Left ensures the resource is of type A.

Implementations

Source

impl<A, B, const TOTAL: u64> Left<A, B, TOTAL>

pub closed spec fn id(self) -> Loc

Returns the unique identifier.

pub closed spec fn protocol_monoid(self) -> SumSP<A, B, TOTAL>

The underlying protocol monoid value for this resource.

pub open spec fn is_resource_owner(self) -> bool

{ self.protocol_monoid().is_resource_owner() }

Whether this token has the right to access the underlying resource.

pub open spec fn resource(self) -> A

{ self.protocol_monoid().resource()->Left_0 }

The resource value, only meaningful if is_resource_owner is true.

pub open spec fn wf(self) -> bool

{
    &&& self.protocol_monoid().is_left()
    &&& self.protocol_monoid().is_valid()
    &&& 1 <= self.frac() <= TOTAL
    &&& self.is_resource_owner() ==> (self.has_resource() <==> !self.has_no_resource())

}

Type invariant.

pub open spec fn has_resource(self) -> bool

{ self.protocol_monoid().has_resource() }

Whether the resource is currently stored, only meaningful if is_resource_owner is true.

pub open spec fn has_no_resource(self) -> bool

{ self.protocol_monoid().has_no_resource() }

Whether the resource has been taken, only meaningful if is_resource_owner is true.

pub open spec fn frac(self) -> int

{ self.protocol_monoid().frac() }

The fraction this token represents.

pub proof fn validate(tracked &self)

ensures
self.wf(),

Left token satisfies the type invariant.

pub proof fn validate_with_left(tracked &mut self, tracked other: &Self)

requires
old(self).id() == other.id(),
ensures
*old(self) == *final(self),
!(final(self).is_resource_owner() && other.is_resource_owner()),
final(self).frac() + other.frac() <= TOTAL,
final(self).wf(),

The existence of two Left tokens with the same id ensures at most one of them is the resource owner.

pub proof fn validate_with_right(tracked &self, tracked other: &Right<A, B, TOTAL>)

ensures
self.id() != other.id(),
self.wf(),

The existence of a Right token ensures they can not have the same id.

pub proof fn tracked_borrow(tracked &self) -> tracked res : &A

requires
self.has_resource(),
ensures
*res == self.resource(),

Borrows the resource of type A.

pub proof fn take_resource(tracked &mut self) -> tracked res : A

requires
old(self).is_resource_owner(),
old(self).has_resource(),
ensures
final(self).id() == old(self).id(),
res == old(self).resource(),
final(self).is_resource_owner(),
final(self).has_no_resource(),
final(self).frac() == old(self).frac(),
final(self).wf(),

Takes the resource out of the token.

pub proof fn put_resource(tracked &mut self, tracked a: A)

requires
old(self).is_resource_owner(),
old(self).has_no_resource(),
ensures
final(self).id() == old(self).id(),
final(self).protocol_monoid()
    == SumSP::<A, B, TOTAL>::Left(Some(a), final(self).frac(), true),
final(self).is_resource_owner(),
final(self).has_resource(),
final(self).resource() == a,
final(self).frac() == old(self).frac(),
final(self).wf(),

Puts a resource of type A back to the token.

pub proof fn split_with_resource(tracked &mut self, n: int) -> tracked res : Self

requires
0 < n < old(self).frac(),
ensures
final(self).id() == old(self).id(),
res.id() == final(self).id(),
final(self).frac() == old(self).frac() - n,
res.frac() == n,
!final(self).is_resource_owner(),
res.is_resource_owner() <==> old(self).is_resource_owner(),
res.is_resource_owner() ==> (res.has_resource() <==> old(self).has_resource()),
res.has_resource() ==> res.resource() == old(self).resource(),
final(self).wf(),
res.wf(),

Splits this token into two Left tokens with the given fraction n, given the resource to the new token if available.

pub proof fn split_without_resource(tracked &mut self, n: int) -> tracked res : Self

requires
0 < n < old(self).frac(),
ensures
final(self).id() == old(self).id(),
res.id() == final(self).id(),
final(self).frac() == old(self).frac() - n,
res.frac() == n,
!res.is_resource_owner(),
final(self).is_resource_owner() <==> old(self).is_resource_owner(),
final(self).is_resource_owner()
    ==> (final(self).has_resource() <==> old(self).has_resource()),
final(self).has_resource() ==> final(self).resource() == old(self).resource(),
final(self).wf(),
res.wf(),

Splits this token into two Left tokens with the given fraction n, without giving the resource to the new token.

pub proof fn update(tracked &mut self, tracked a: A) -> tracked res : Option<A>

requires
old(self).is_resource_owner(),
ensures
final(self).id() == old(self).id(),
final(self).is_resource_owner(),
final(self).has_resource(),
final(self).resource() == a,
final(self).frac() == old(self).frac(),
res == if old(self).has_resource() { Some(old(self).resource()) } else { None },
final(self).wf(),

Updates the token with a new resource of type A, and returns the old resource if available.