OneRightKnowledge

Struct OneRightKnowledge 

pub struct OneRightKnowledge<A, B, const TOTAL: u64 = 2> { /* private fields */ }
OneRightKnowledge is a special case of Right that the fraction is always one and it does not own the resource.

Implementations

Source

impl<A, B, const TOTAL: u64> OneRightKnowledge<A, B, TOTAL>

pub closed spec fn id(self) -> Loc

Returns the unique identifier.

pub closed spec fn protocol_monoid(self) -> CsumP<A, B, TOTAL>

The underlying protocol monoid value for this resource.

pub closed spec fn frac(self) -> int

Returns the fraction of this token, which should always be 1.

pub open spec fn wf(self) -> bool

{
    &&& self.protocol_monoid().is_right()
    &&& self.protocol_monoid().is_valid()
    &&& !self.protocol_monoid().is_resource_owner()
    &&& self.frac() == 1
    &&& TOTAL >= 1

}

Type invariant.

pub proof fn validate(tracked &self)

ensures
self.wf(),

OneRightKnowledge token satisfies the type invariant.

pub proof fn validate_with_one_left_knowledge(tracked &self, tracked other: &OneLeftKnowledge<A, B, TOTAL>, )

ensures
self.id() != other.id(),

The existence of OneLeftKnowledge token ensures they can not have the same id.

pub proof fn validate_with_one_left_owner(tracked &self, tracked other: &OneLeftOwner<A, B, TOTAL>)

ensures
self.id() != other.id(),

The existence of OneLeftOwner token ensures they can not have the same id.