OneRightOwner

Struct OneRightOwner 

pub struct OneRightOwner<A, B, const TOTAL: u64 = 2> { /* private fields */ }
OneRightOwner is a special case of Right that the fraction is always one and it is the resource owner.

Implementations

Source

impl<A, B, const TOTAL: u64> OneRightOwner<A, B, TOTAL>

pub closed spec fn id(self) -> Loc

Returns the unique identifier.

pub closed spec fn protocol_monoid(self) -> SumSP<A, B, TOTAL>

The underlying protocol monoid value for this resource.

pub closed spec fn resource(self) -> B

Returns the resource owned by this token.

pub closed spec fn frac(self) -> int

Returns the fraction of this token, which should always be 1.

pub closed spec fn has_resource(self) -> bool

Whether this token currently has the resource stored.

pub closed spec fn has_no_resource(self) -> bool

Whether the resource has been taken from this token.

pub open spec fn wf(self) -> bool

{
    &&& self.protocol_monoid().is_right()
    &&& self.protocol_monoid().is_valid()
    &&& self.protocol_monoid().is_resource_owner()
    &&& self.frac() == 1
    &&& TOTAL >= 1
    &&& self.has_resource() <==> !self.has_no_resource()

}

Type invariant.

pub proof fn validate(tracked &self)

ensures
self.wf(),

OneRightOwner token satisfies the type invariant.

pub proof fn validate_with_one_right_owner(tracked &mut self, tracked other: &Self)

ensures
*old(self) == *final(self),
final(self).id() != other.id(),
final(self).wf(),

The existence of two OneRightOwner tokens ensures they can not have the same id.

pub proof fn tracked_borrow(tracked &self) -> tracked res : &B

requires
self.has_resource(),
ensures
*res == self.resource(),

Borrows the resource of type B.

pub proof fn take_resource(tracked &mut self) -> tracked res : B

requires
old(self).has_resource(),
ensures
res == old(self).resource(),
final(self).has_no_resource(),
final(self).wf(),

Takes the resource out of the token.

pub proof fn put_resource(tracked &mut self, tracked b: B)

requires
old(self).has_no_resource(),
ensures
final(self).resource() == b,
final(self).has_resource(),
final(self).wf(),

Puts a resource of type B back to the token.

pub proof fn update(tracked &mut self, tracked b: B) -> tracked res : Option<B>

ensures
final(self).resource() == b,
final(self).has_resource(),
final(self).wf(),
res == if old(self).has_resource() { Some(old(self).resource()) } else { None },

Updates the resource of type B in the token, and returns the old resource if available.