pub struct Right<A, B, const TOTAL: u64 = 2> { /* private fields */ }Expand description
Right ensures the resource is of type B.
Implementations
Sourceimpl<A, B, const TOTAL: u64> Right<A, B, TOTAL>
impl<A, B, const TOTAL: u64> Right<A, B, TOTAL>
Sourcepub closed spec fn protocol_monoid(self) -> SumSP<A, B, TOTAL>
pub closed spec fn protocol_monoid(self) -> SumSP<A, B, TOTAL>
The underlying protocol monoid value for this resource.
Sourcepub open spec fn is_resource_owner(self) -> bool
pub open spec fn is_resource_owner(self) -> bool
{ self.protocol_monoid().is_resource_owner() }Whether this token has the right to access the underlying resource.
Sourcepub open spec fn resource(self) -> B
pub open spec fn resource(self) -> B
{ self.protocol_monoid().resource()->Right_0 }The resource value, only meaningful if is_resource_owner is true.
Sourcepub open spec fn wf(self) -> bool
pub open spec fn wf(self) -> bool
{
&&& self.protocol_monoid().is_right()
&&& self.protocol_monoid().is_valid()
&&& 1 <= self.frac() <= TOTAL
&&& self.is_resource_owner() ==> (self.has_resource() <==> !self.has_no_resource())
}Type invariant.
Sourcepub open spec fn has_resource(self) -> bool
pub open spec fn has_resource(self) -> bool
{ self.protocol_monoid().has_resource() }Whether the resource is currently stored, only meaningful if is_resource_owner is true.
Sourcepub open spec fn has_no_resource(self) -> bool
pub open spec fn has_no_resource(self) -> bool
{ self.protocol_monoid().has_no_resource() }Whether the resource has been taken, only meaningful if is_resource_owner is true.
Sourcepub open spec fn frac(self) -> int
pub open spec fn frac(self) -> int
{ self.protocol_monoid().frac() }The fraction this token represents.
Sourcepub proof fn validate(tracked &self)
pub proof fn validate(tracked &self)
self.wf(),Right token satisfies the type invariant.
Sourcepub proof fn validate_with_left(tracked &self, tracked other: &Left<A, B, TOTAL>)
pub proof fn validate_with_left(tracked &self, tracked other: &Left<A, B, TOTAL>)
self.id() != other.id(),The existence of a Left token ensures they can not have the same id.
Sourcepub proof fn validate_with_right(tracked &mut self, tracked other: &Self)
pub proof fn validate_with_right(tracked &mut self, tracked other: &Self)
old(self).id() == other.id(),ensures*old(self) == *final(self),!(final(self).is_resource_owner() && other.is_resource_owner()),final(self).frac() + other.frac() <= TOTAL,final(self).wf(),The existence of two Right tokens with the same id ensures at most one of them is the resource owner.
Sourcepub proof fn tracked_borrow(tracked &self) -> tracked res : &B
pub proof fn tracked_borrow(tracked &self) -> tracked res : &B
self.has_resource(),ensures*res == self.resource(),Borrows the resource of type B.
Sourcepub proof fn take_resource(tracked &mut self) -> tracked res : B
pub proof fn take_resource(tracked &mut self) -> tracked res : B
old(self).is_resource_owner(),old(self).has_resource(),ensuresfinal(self).id() == old(self).id(),res == old(self).resource(),final(self).is_resource_owner(),final(self).has_no_resource(),final(self).frac() == old(self).frac(),final(self).wf(),Takes the resource out of the token.
Sourcepub proof fn put_resource(tracked &mut self, tracked b: B)
pub proof fn put_resource(tracked &mut self, tracked b: B)
old(self).is_resource_owner(),old(self).has_no_resource(),ensuresfinal(self).id() == old(self).id(),final(self).protocol_monoid()
== SumSP::<A, B, TOTAL>::Right(Some(b), final(self).frac(), true),final(self).is_resource_owner(),final(self).has_resource(),final(self).resource() == b,final(self).frac() == old(self).frac(),final(self).wf(),Puts a resource of type B back to the token.
Sourcepub proof fn split_with_resource(tracked &mut self, n: int) -> tracked res : Self
pub proof fn split_with_resource(tracked &mut self, n: int) -> tracked res : Self
0 < n < old(self).frac(),ensuresfinal(self).id() == old(self).id(),res.id() == final(self).id(),final(self).frac() == old(self).frac() - n,res.frac() == n,!final(self).is_resource_owner(),res.is_resource_owner() <==> old(self).is_resource_owner(),res.is_resource_owner() ==> (res.has_resource() <==> old(self).has_resource()),res.has_resource() ==> res.resource() == old(self).resource(),final(self).wf(),res.wf(),Splits this token into two Right tokens with the given fraction n, given the resource to the new token if available.
Sourcepub proof fn split_without_resource(tracked &mut self, n: int) -> tracked res : Self
pub proof fn split_without_resource(tracked &mut self, n: int) -> tracked res : Self
0 < n < old(self).frac(),ensuresfinal(self).id() == old(self).id(),res.id() == final(self).id(),final(self).frac() == old(self).frac() - n,res.frac() == n,!res.is_resource_owner(),final(self).is_resource_owner() <==> old(self).is_resource_owner(),final(self).is_resource_owner()
==> (final(self).has_resource() <==> old(self).has_resource()),final(self).has_resource() ==> final(self).resource() == old(self).resource(),final(self).wf(),res.wf(),Splits this token into two Right tokens with the given fraction n, without giving the resource to the new token.
Sourcepub proof fn update(tracked &mut self, tracked b: B) -> tracked res : Option<B>
pub proof fn update(tracked &mut self, tracked b: B) -> tracked res : Option<B>
old(self).is_resource_owner(),ensuresfinal(self).id() == old(self).id(),final(self).is_resource_owner(),final(self).has_resource(),final(self).resource() == b,final(self).frac() == old(self).frac(),res == if old(self).has_resource() { Some(old(self).resource()) } else { None },final(self).wf(),Updates the token with a new resource of type B, and returns the old resource if available.